
#ifdef HASKELL_ORB_MICO
#define MICO_CONF_DYNANY
#define MICO_CONF_POA
#include <CORBA-SMALL.h>
#endif

#ifdef HASKELL_ORB_ORBACUS
#include <OB/CORBA.h>
#endif

#include <map>

#include "show_corba.h"

// We want C linking for all functions callable from Haskell
extern "C" {

#define ARRIVED cerr << "Arrived in " << __FILE__ << " at " << __LINE__ << endl; flush(cerr);

#ifdef HASKELL_ORB_MICO
#if MICO_BIN_VERSION < 0x020305
#define NON_CONST_ANY_EXTRACTION
#endif
#endif

////////////////////////////////
//        GLOBALS
////////////////////////////////

// Set by 'corba_orb_init'
static CORBA::ORB_ptr the_orb;

static PortableServer::POA_ptr the_poa;

static DynamicAny::DynAnyFactory_var dynany_factory;

// Useful for debugging
static CORBA_OUT out(cout);

////////////////////////////////
//       UTILITY FUNCTIONS
////////////////////////////////

void set_type_of_any(CORBA::Any * any, CORBA::TypeCode_ptr tc)
{
#ifdef HASKELL_ORB_MICO

#if MICO_BIN_VERSION >= 0x020305
  any->set_type(tc);
#else
  any->type(tc);
#endif

#else

  // ORBacus
  any->replace(tc, 0);

#endif
}

////////////////////////////////
//      CStringArray
////////////////////////////////

typedef char ** CStringArray;
  
CStringArray
c_string_array_create(int size)
{
  return new char*[size];
}

void 
c_string_array_set(CStringArray arr, int index, char * value)			
{
  arr[index] = strdup(value);
}

char * 
c_string_array_get(CStringArray arr, 
				   int index)
{
  return strdup(arr[index]);
}

void
c_string_array_delete(CStringArray arr)
{
  delete[] arr;
}

////////////////////////////////
//          ORB
////////////////////////////////

CORBA::ORB_ptr
corba_orb_init(CStringArray argv, int argc,
			   CStringArray * out_argv, int * out_argc)
{
  // Save for further use
  //
  // This makes it unnecessary to pass the orb as an argument to
  // functions which need to call ORB related functions, like
  // create_dyn_struct().
  //
  the_orb = CORBA::ORB_init(argc, argv, "mico-local-orb");

  CORBA::Object_var obj =
    the_orb->resolve_initial_references ("DynAnyFactory");
  
  dynany_factory = DynamicAny::DynAnyFactory::_narrow (obj);
  assert(dynany_factory != (void*)NULL);

  obj = the_orb->resolve_initial_references("RootPOA");
  the_poa = PortableServer::POA::_narrow(obj);
  assert(the_poa);

  out.set_orb(the_orb);
  out.set_fac(dynany_factory);

  *out_argv = argv;
  *out_argc = argc;
  
  return the_orb;
}

CORBA::Object_ptr
corba_orb_string_to_object(CORBA::ORB_ptr orb, 
						   char * str)
{
  CORBA::Object_ptr result;

  // ORBacus does not like the newline at the end of an IOR...
  if ((strlen(str) > 0) && (str[strlen(str) - 1] == '\n'))
	str[strlen(str) - 1] = 0;

  try {
	result = orb->string_to_object(str);
  }
  catch (const CORBA::Exception& ex) {
	cout << ex << endl;
	result = NULL;
  }

  return result;
}

CORBA::Object_ptr
corba_orb_bind(CORBA::ORB_ptr orb, 
			   char * repoid, 
			   char * address)
{
  CORBA::Object_ptr result;

#ifdef HASKELL_ORB_MICO
  result = orb->bind(repoid, address);
  if (CORBA::is_nil(result)) {
    result = NULL;
  }
#else
  // Bind is MICO specific
  result = NULL;
#endif

  return result;
}

char *
corba_orb_object_to_string(CORBA::ORB_ptr orb,
						   CORBA::Object_ptr obj)
{
  try {
	return strdup(orb->object_to_string(obj));
  }
  catch (...) {
	return strdup("");
  }
}

CORBA::Object_ptr
corba_orb_resolve_initial_references(CORBA::ORB_ptr orb,
									 char * name)
{
  CORBA::Object_ptr result;
  try {
	result = orb->resolve_initial_references(name);
  }
  catch (...) {
	result = NULL;
  }

  return result;
}

////////////////////////////////
//          OBJECT
////////////////////////////////

CORBA::Request_ptr
corba_object_create_request(CORBA::Object_ptr obj,
							char * operation)
{
  return obj->_request(operation);
}

bool
corba_object_is_a(CORBA::Object_ptr obj,
				  char * repoid)
{
  try {
	return obj->_is_a(repoid);
  }
  catch (...) {
	return false;
  }
}

CORBA::Object_ptr
corba_object_nil()
{
  return CORBA::Object::_nil();
}

int
corba_object_hash(CORBA::Object_ptr obj,
				  CORBA::ULong max)
{
  return obj->_hash(max);
}

////////////////////////////////
//         TYPECODE
////////////////////////////////

CORBA::StructMemberSeq * 
corba_struct_member_seq_create(int size)
{
  CORBA::StructMemberSeq * result = new CORBA::StructMemberSeq();
  result->length(size);
  return result;
}

void
corba_struct_member_seq_fill(CORBA::StructMemberSeq * seq,
							 int index,
							 char * name,
							 CORBA::TypeCode_ptr type)
{
  (*seq)[index].name = CORBA::string_dup(name);
  (*seq)[index].type = type;
  (*seq)[index].type_def = CORBA::IDLType::_nil();
}

void
corba_struct_member_seq_delete(CORBA::StructMemberSeq * seq)
{
  delete seq;
}

CORBA::UnionMemberSeq *
corba_union_member_seq_create(int size)
{
  CORBA::UnionMemberSeq * result = new CORBA::UnionMemberSeq();
  result->length(size);
  return result;
}

void
corba_union_member_seq_fill(CORBA::UnionMemberSeq * seq,
							int index,
							char * name,
							CORBA::Any * label,
							CORBA::TypeCode_ptr type)
{
  (*seq)[index].name = CORBA::string_dup(name);
  (*seq)[index].label = *label;
  (*seq)[index].type = type;
}

void
corba_union_member_seq_delete(CORBA::UnionMemberSeq * seq)
{
  delete seq;
}

CORBA::EnumMemberSeq *
corba_enum_member_seq_create(int size)
{
  CORBA::EnumMemberSeq * result = new CORBA::EnumMemberSeq();
  result->length(size);
  return result;
}

void
corba_enum_member_seq_fill(CORBA::EnumMemberSeq * seq,
						   int index,
						   char * name)
{
  (*seq)[index] = CORBA::string_dup(name);
}

void
corba_enum_member_seq_delete(CORBA::EnumMemberSeq * seq)
{
  delete seq;
}

CORBA::TypeCode_ptr
corba_tc_create_basic(CORBA::TCKind kind)
{
#ifdef HASKELL_ORB_ORBACUS
  CORBA::TypeCode * tmp_tc;

  // ORBacus does not have a simple typecode creation function :(
  switch (kind) {
  case CORBA::tk_null:
	tmp_tc = CORBA::_tc_null;
	break;
  case CORBA::tk_void:
	tmp_tc = CORBA::_tc_void;
	break;

  case CORBA::tk_short: tmp_tc = CORBA::_tc_short; break;
  case CORBA::tk_long: tmp_tc = CORBA::_tc_long; break;
  case CORBA::tk_longlong: tmp_tc = CORBA::_tc_longlong; break;
  case CORBA::tk_ushort: tmp_tc = CORBA::_tc_ushort; break;
  case CORBA::tk_ulong: tmp_tc = CORBA::_tc_ulong; break;
  case CORBA::tk_ulonglong: tmp_tc = CORBA::_tc_ulonglong; break;
  case CORBA::tk_float: tmp_tc = CORBA::_tc_float; break;
  case CORBA::tk_double: tmp_tc = CORBA::_tc_double; break;
  case CORBA::tk_longdouble: tmp_tc = CORBA::_tc_longdouble; break;
  case CORBA::tk_boolean: tmp_tc = CORBA::_tc_boolean; break;
  case CORBA::tk_char: tmp_tc = CORBA::_tc_char; break;
  case CORBA::tk_wchar: tmp_tc = CORBA::_tc_wchar; break;
  case CORBA::tk_octet: tmp_tc = CORBA::_tc_octet; break;
  case CORBA::tk_any: tmp_tc = CORBA::_tc_any; break;
  case CORBA::tk_TypeCode: tmp_tc = CORBA::_tc_TypeCode; break;
  case CORBA::tk_Principal: tmp_tc = CORBA::_tc_Principal; break;
  default:
	assert(0);
  }

  return CORBA::TypeCode::_duplicate(tmp_tc);
#else
  try {
	return CORBA::TypeCode::create_basic_tc(kind);
  } 
  catch (...) {
	cout << "Error: create_tc_basic(" << (int)kind << ") failed." << endl;
	assert(0);
  }
#endif
}

CORBA::TypeCode_ptr
corba_tc_create_string(int bound)
{
  try {
	return the_orb->create_string_tc(bound);
  }
  catch (...) {
	cout << "Error: create_tc_string(" << bound << ") failed." << endl;
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_objref(char * repoid, 
					   char * name)
{
  try {
	return the_orb->create_interface_tc(repoid, name);
  }
  catch (...) {
	cout << "Error: create_interface_tc(" << repoid << ", " << name << ") failed." << endl;
	assert(0);
  }	
}

CORBA::TypeCode_ptr
corba_tc_create_struct(char * repoid, 
					   char * name,
					   CORBA::StructMemberSeq * members)
{
  try {
	return the_orb->create_struct_tc(repoid, name, *members);
  }
  catch (const CORBA::Exception& ex) {
	cout << ex << endl;
	cout << "Error: create_struct_tc(" << repoid << ", " << name << ") failed." << endl;
	for (unsigned int i = 0; i < members->length(); ++i) {
	  cout << "Member " << i << ": " << (*members)[i].name << endl;
	  out.display_IDL_TypeCode((*members)[i].type);
	}
	assert(0);
  }	
}

CORBA::TypeCode_ptr
corba_tc_create_union(char * repoid, 
					  char * name,
					  CORBA::TypeCode_ptr label_type,
					  CORBA::UnionMemberSeq * members)
{
  try {
	return the_orb->create_union_tc(repoid, name, label_type, *members);
  }
  catch (...) {
	cout << "Error: create_union_tc(" << repoid << ", " << name << ") failed." << endl;
	cout << "Label type: " << endl;
	out.display_IDL_TypeCode(label_type);
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_sequence(int bound,
						 CORBA::TypeCode_ptr content_type)
{
  try {
	return the_orb->create_sequence_tc(bound, content_type);
  }
  catch (...) {
	cout << "Error: create_sequence_tc() failed." << endl;
	cout << "Content type: " << endl;
	out.display_IDL_TypeCode(content_type);
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_enum(char * repoid, 
					 char * name,
					 CORBA::EnumMemberSeq * members)
{
  try {
	return the_orb->create_enum_tc(repoid, name, *members);
  }
  catch (...) {
	cout << "Error: create_enum_tc() failed." << endl;
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_alias(char * repoid, 
					  char * name,
					  CORBA::TypeCode_ptr content_type)
{
  try {
	return the_orb->create_alias_tc(repoid, name, content_type);
  }
  catch (...) {
	cout << "Error: create_alias_tc() failed." << endl;
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_array(CORBA::TypeCode_ptr content_type,
					  int length)
{
  try {
	return the_orb->create_array_tc(length, content_type);
  }
  catch (...) {
	cout << "Error: create_array_tc() failed." << endl;
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_exception(char * repoid, 
						  char * name,
						  CORBA::StructMemberSeq * members)
{
  try {
	return the_orb->create_exception_tc(repoid, name, *members);
  }
  catch (...) {
	cout << "Error: create_exception_tc() failed." << endl;
	assert(0);
  }
}

CORBA::TypeCode_ptr
corba_tc_create_recursive(char * repoid) 
{
  try {
	return the_orb->create_recursive_tc(repoid);
  }
  catch (...) {
	cout << "Error: create_recursive_tc() failed." << endl;
	assert(0);
  }
}

void
corba_tc_release(CORBA::TypeCode_ptr tc)
{
  CORBA::release(tc);
}

// Accessor functions

int
corba_tc_get_kind(CORBA::TypeCode_ptr tc)
{
  return tc->kind();
}

char *
corba_tc_get_id(CORBA::TypeCode_ptr tc)
{
  return strdup(tc->id());
}

char *
corba_tc_get_name(CORBA::TypeCode_ptr tc)
{
  return strdup(tc->name());
}

int
corba_tc_get_member_count(CORBA::TypeCode_ptr tc)
{
  return tc->member_count();
}

char *
corba_tc_get_member_name(CORBA::TypeCode_ptr tc,
						 int index)
{
  return strdup(tc->member_name(index));
}

CORBA::TypeCode_ptr
corba_tc_get_member_type(CORBA::TypeCode_ptr tc,
						 int index)
{
  return tc->member_type(index);
}

CORBA::Any *
corba_tc_get_member_label(CORBA::TypeCode_ptr tc,
						  int index)
{
  return tc->member_label(index);
}

CORBA::TypeCode_ptr 
corba_tc_get_discriminator_type(CORBA::TypeCode_ptr tc)
{
  return tc->discriminator_type();
}

int
corba_tc_get_default_index(CORBA::TypeCode_ptr tc)
{
  return tc->default_index();
}

int
corba_tc_get_length(CORBA::TypeCode_ptr tc)
{
  return tc->length();
}

CORBA::TypeCode_ptr
corba_tc_get_content_type(CORBA::TypeCode_ptr tc)
{
  return tc->content_type();
}

////////////////////////////////
//        ANY
////////////////////////////////

CORBA::Any * 
corba_any_create_null()
{
  // The default constructed Any has a type code of tc_null
  return new CORBA::Any;
}

CORBA::Any * 
corba_any_create_short(int value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::Short>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_ushort(int value)
{
  CORBA::Any *out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::UShort>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_long(int value)
{
  CORBA::Any *out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::Long>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_ulong(int value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::ULong>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_float(double value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::Float>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_double(double value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= static_cast<CORBA::Double>(value);
  return out_any;
}

CORBA::Any * 
corba_any_create_bool(int value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= CORBA::Any::from_boolean(static_cast<CORBA::Boolean>(value));
  return out_any;
}

CORBA::Any * 
corba_any_create_char(int value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= CORBA::Any::from_char(static_cast<CORBA::Char>(value));
  return out_any;
}

CORBA::Any * 
corba_any_create_octet(int value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= CORBA::Any::from_octet(static_cast<CORBA::Octet>(value));
  return out_any;
}

CORBA::Any * 
corba_any_create_string(char * value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= strdup(value);

  return out_any;
}

CORBA::Any * 
corba_any_create_any(CORBA::Any * value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= value;
  return out_any;
}

CORBA::Any * 
corba_any_create_tc(CORBA::TypeCode_ptr value)
{
  CORBA::Any * out_any = new CORBA::Any;
  *out_any <<= value;
  return out_any;
}

CORBA::Any * 
corba_any_create_objref(CORBA::Object_ptr value)
{
  CORBA::Any * out_any = new CORBA::Any;

  *out_any <<= value;

  return out_any;
  // MICO Does not provide <<= for Object_ptr ???
  //  *out_any <<= CORBA::Any::from_object(value, "Object");
}

void
corba_any_delete(CORBA::Any * any)
{
  delete any;
}


CORBA::TypeCode_ptr
corba_any_get_type(CORBA::Any * any)
{
  return any->type();
}

int
corba_any_extract_short(CORBA::Any * any)
{
  CORBA::Short tmp;

  assert( *(any) >>= tmp );

  return tmp;
}

int
corba_any_extract_long(CORBA::Any * any)
{
  CORBA::Long tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

int
corba_any_extract_ushort(CORBA::Any * any)
{
  CORBA::UShort tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

int
corba_any_extract_ulong(CORBA::Any * any)
{
  CORBA::ULong tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

double
corba_any_extract_float(CORBA::Any * any)
{
  CORBA::Float tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

double
corba_any_extract_double(CORBA::Any * any)
{
  CORBA::Double tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

bool
corba_any_extract_boolean(CORBA::Any * any)
{
  CORBA::Boolean tmp;

  assert( *(any) >>= CORBA::Any::to_boolean(tmp) );
  return tmp;
}

char
corba_any_extract_char(CORBA::Any * any)
{
  CORBA::Char tmp;

  assert( *(any) >>= CORBA::Any::to_char(tmp) );
  return tmp;
}

int
corba_any_extract_octet(CORBA::Any * any)
{
  CORBA::Octet tmp;

  assert( *(any) >>= CORBA::Any::to_octet(tmp) );
  return tmp;
}

char *
corba_any_extract_string(CORBA::Any * any)
{
#ifdef NON_CONST_ANY_EXTRACTION
  char * tmp;
#else
  const char * tmp;
#endif

  assert( *(any) >>= tmp );
  return strdup(tmp);
}

const CORBA::Any *
corba_any_extract_any(CORBA::Any * any)
{
#ifdef NON_CONST_ANY_EXTRACTION
  CORBA::Any * value;
#else
  const CORBA::Any * value;
#endif

  assert( *(any) >>= value );
  return value;
}

CORBA::TypeCode_ptr
corba_any_extract_tc(CORBA::Any * any)
{
  CORBA::TypeCode_ptr tmp;

  assert( *(any) >>= tmp );
  return tmp;
}

CORBA::Object_ptr
corba_any_extract_objref(CORBA::Any * any)
{
  CORBA::Object_ptr value;

  assert( *(any) >>= CORBA::Any::to_object(value) );
  return CORBA::Object::_duplicate(value);
}

////////////////////////////////
//        DYNANY
////////////////////////////////

DynamicAny::NameValuePairSeq *
corba_nv_pair_seq_create(int size)
{
  DynamicAny::NameValuePairSeq * out_value = new DynamicAny::NameValuePairSeq();
  out_value->length(size);
  return out_value;
}

void
corba_nv_pair_seq_fill(DynamicAny::NameValuePairSeq * seq,
					   int index,
					   char * name,
					   CORBA::Any * value)
{
  (*seq)[index].id = CORBA::string_dup(name);
  (*seq)[index].value = *value;
}

int
corba_nv_pair_seq_length(DynamicAny::NameValuePairSeq * seq)
{
  return seq->length();
}

void
corba_nv_pair_seq_get(DynamicAny::NameValuePairSeq * seq,
					  int index,
					  char * * out_name,
					  CORBA::Any ** out_value)
{
  *out_name = strdup((*seq)[index].id);
  *out_value = &((*seq)[index].value);
}

void
corba_nv_pair_seq_delete(DynamicAny::NameValuePairSeq * seq)
{
  delete seq;
}

CORBA::Any *
corba_dyn_struct_create(CORBA::TypeCode_ptr tc,
						DynamicAny::NameValuePairSeq * members)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynStruct_var ds =
    DynamicAny::DynStruct::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  /* FOR TESTING:
  DynamicAny::NameValuePairSeq  * components = ds->get_members();

  for (unsigned int i = 0; i < members->length(); ++i) {
	if (! ((*members)[i].value.type())->equivalent((*components)[i].value.type())) {
	  cout << (*members)[i].value.type()->id() << endl;
	  cout << (*components)[i].value.type()->id() << endl;
	  //	  out.display_IDL_TypeCode(((*members)[i]).value.type());
	  //	  out.display_IDL_TypeCode(((*components)[i]).value.type());
	}
  }
  */

  try {
	ds->set_members(*members);
  }
  catch (...) {
	cout << "Error: DynStruct::set_members() failed." << endl;
	assert(0);
  }

  CORBA::Any * out_any = ds->to_any();

  ds->destroy();
  return out_any;
}

DynamicAny::NameValuePairSeq *
corba_dyn_struct_members(CORBA::Any * any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynStruct_var ds =
    DynamicAny::DynStruct::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  DynamicAny::NameValuePairSeq * out_members = ds->get_members();

  ds->destroy();
  return out_members;
}

DynamicAny::AnySeq *
corba_any_seq_create(int size)
{
  DynamicAny::AnySeq * out_value = new DynamicAny::AnySeq();
  out_value->length(size);
  return out_value;
}

void
corba_any_seq_fill(DynamicAny::AnySeq * seq,
				   int index,
				   CORBA::Any * value)
{
  (*seq)[index] = *value;
}

int
corba_any_seq_length(DynamicAny::AnySeq * seq)
{
  return seq->length();
}

CORBA::Any *
corba_any_seq_get(DynamicAny::AnySeq * seq,
				  int index)
{
  return &((*seq)[index]);
}

void
corba_any_seq_delete(DynamicAny::AnySeq * seq)
{
  delete seq;
}

CORBA::Any *
corba_dyn_sequence_create(CORBA::TypeCode_ptr tc,
						  DynamicAny::AnySeq * elements)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynSequence_var ds =
    DynamicAny::DynSequence::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  ds->set_length(elements->length());
  ds->set_elements(*elements);
  CORBA::Any * out_any = ds->to_any();

  ds->destroy();
  return out_any;
}

DynamicAny::AnySeq *
corba_dyn_sequence_elements(CORBA::Any * any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynSequence_var ds =
    DynamicAny::DynSequence::_narrow (tmp);
  assert(!CORBA::is_nil(ds));

  DynamicAny::AnySeq * seq = ds->get_elements();

  ds->destroy();
  return seq;
}

CORBA::Any *
corba_dyn_union_create(CORBA::TypeCode_ptr tc,
					   CORBA::Any * label,
					   CORBA::Any * value)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code(tc);
  DynamicAny::DynUnion_var du =
    DynamicAny::DynUnion::_narrow (tmp);
  assert(!CORBA::is_nil(du));

  du->set_discriminator(dynany_factory->create_dyn_any(*label));
  if (! (du->has_no_active_member())) {
	// ORBacus needs this...
	du->rewind();
	du->next();
	du->current_component()->from_any(*value);
  }

  CORBA::Any * out_any = du->to_any();

  du->destroy();
  return out_any;
}

void
corba_dyn_union_get(CORBA::Any * any,
					CORBA::Any ** out_label,
					CORBA::Any ** out_member)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynUnion_var du =
    DynamicAny::DynUnion::_narrow (tmp);
  assert(!CORBA::is_nil(du));

  *out_label = du->get_discriminator()->to_any();
  if (du->has_no_active_member()) {
	// There must be a better way to create an Any...
	// This any has type code tk_null
	*out_member = new CORBA::Any;
  }
  else {
	*out_member = du->member()->to_any();
  }

  du->destroy();
}

CORBA::Any *
corba_dyn_enum_create(CORBA::TypeCode_ptr tc,
					  char * value)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code(tc);
  DynamicAny::DynEnum_var de =
    DynamicAny::DynEnum::_narrow (tmp);
  assert(!CORBA::is_nil(de));

  de->set_as_string(strdup(value));

  CORBA::Any * out_any = de->to_any();

  de->destroy();
  return out_any;
}

char *
corba_dyn_enum_get(CORBA::Any * any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynEnum_var de =
    DynamicAny::DynEnum::_narrow (tmp);
  assert(!CORBA::is_nil(de));

  char * out_value = strdup(de->get_as_string());

  de->destroy();
  return out_value;
}

CORBA::Any *
corba_dyn_array_create(CORBA::TypeCode_ptr tc,
					   DynamicAny::AnySeq * elements)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any_from_type_code (tc);
  DynamicAny::DynArray_var da =
    DynamicAny::DynArray::_narrow (tmp);
  assert(!CORBA::is_nil(da));

  da->set_elements(*elements);
  CORBA::Any * out_any = da->to_any();

  da->destroy();
  return out_any;
}

DynamicAny::AnySeq *
corba_dyn_array_elements(CORBA::Any * any)
{
  assert(the_orb);

  DynamicAny::DynAny_var tmp =
    dynany_factory->create_dyn_any(*any);
  DynamicAny::DynArray_var da =
    DynamicAny::DynArray::_narrow (tmp);
  assert(!CORBA::is_nil(da));

  DynamicAny::AnySeq * out_elements = da->get_elements();

  da->destroy();
  return out_elements;
}

////////////////////////////////
//       NAMEDVALUE
////////////////////////////////

char *
corba_nv_get_name(CORBA::NamedValue_ptr nv)
{
  return strdup(nv->name());
}

CORBA::Any *
corba_nv_get_value(CORBA::NamedValue_ptr nv)
{
  return nv->value();
}

int
corba_nv_get_flags(CORBA::NamedValue_ptr nv)
{
  return nv->flags();
}

void
corba_nv_set_type(CORBA::NamedValue_ptr nv,
				  CORBA::TypeCode_ptr tc)
{
  set_type_of_any(nv->value(), tc);
}

void
corba_nv_set_value(CORBA::NamedValue_ptr nv,
				   CORBA::Any * value)
{
  (*nv->value()) = *value;
}

////////////////////////////////
//           NVLIST
////////////////////////////////

CORBA::NVList_ptr
corba_nvlist_create()
{
  CORBA::NVList_ptr out_nvlist;
  the_orb->create_list(0, out_nvlist);
  return out_nvlist;
}

CORBA::NamedValue_ptr
corba_nvlist_add(CORBA::NVList_ptr nvlist,
				 int flags)
{
  return nvlist->add(flags);
}

int
corba_nvlist_count(CORBA::NVList_ptr nvlist)
{
  return nvlist->count();
}

CORBA::NamedValue_ptr
corba_nvlist_get(CORBA::NVList_ptr nvlist,
				 int index)
{
  return nvlist->item(index);
}

////////////////////////////////
//       REQUEST
////////////////////////////////

CORBA::NVList_ptr
corba_request_arguments(CORBA::Request_ptr req)
{
  return req->arguments();
}

void 
corba_request_add_in_arg(CORBA::Request_ptr req,
						 char * name,
						 CORBA::Any * value)
{
  char * tmp = strdup(name);

  (req->add_in_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_out_arg(CORBA::Request_ptr req,
						  char * name,
						  CORBA::Any * value)
{
  char * tmp = strdup(name);

  (req->add_out_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_inout_arg(CORBA::Request_ptr req,
							char * name,
							CORBA::Any * value)
{
  char * tmp = strdup(name);

  (req->add_inout_arg(tmp)) = *value;
  delete[] tmp;
}

void 
corba_request_add_out_arg_template(CORBA::Request_ptr req,
								   char * name,
								   CORBA::TypeCode_ptr tc)
{
  char * tmp = strdup(name);

  set_type_of_any(&(req->add_out_arg(tmp)), tc);

  delete[] tmp;
}

void 
corba_request_invoke(CORBA::Request_ptr req)
{
#ifdef HASKELL_ORB_MICO
  // Apparently, if a SystemException occurs during the invoke() call,
  // MICO raises it as a C++ exception instead of putting it into the
  // environment.
  // TODO: Is this still true in 2.3.5?
  try {
	req->invoke();
  }
  catch (const CORBA::SystemException& e) {
	req->env()->exception(new CORBA::SystemException(e));
  }
#else
	req->invoke();
#endif
}

void 
corba_request_set_result_tc(CORBA::Request_ptr req,
							CORBA::TypeCode_ptr tc)
{
  set_type_of_any(req->result()->value(), tc);
}
  
CORBA::Any *
corba_request_result(CORBA::Request_ptr req)
{
  return req->result()->value();
}

CORBA::Exception *
corba_request_exception(CORBA::Request_ptr req)
{
  return req->env()->exception();
}

void
corba_request_add_exception_def(CORBA::Request_ptr req,
								CORBA::TypeCode_ptr tc)
{
  //  out.display_IDL_TypeCode(tc);
  req->exceptions()->add(tc);
}

void
corba_request_delete(CORBA::Request_ptr req)
{
#ifdef HASKELL_ORB_MICO
  // TODO: Is this right?
  delete req;
#else
  release(req);
#endif
}

////////////////////////////////
//    EXCEPTION
////////////////////////////////

int
corba_exception_is_system(CORBA::Exception * ex)
{
  // For some reason, dynamic_cast<> crashes on NT with VC6 and MICO 2.3.5 :(
  return (CORBA::SystemException::_downcast(ex) != 0);
}

int
corba_exception_get_minor(CORBA::Exception * ex)
{
  CORBA::SystemException * se = CORBA::SystemException::_downcast(ex);
  assert(se != NULL);
  return se->minor();
}

int
corba_exception_get_completed(CORBA::Exception * ex)
{
  CORBA::SystemException * se = CORBA::SystemException::_downcast(ex);
  assert(se != NULL);
  return se->completed();
}

char *
corba_exception_get_repoid(CORBA::Exception * ex)
{
#ifdef HASKELL_ORB_MICO
  return strdup(ex->_repoid());
#else
  return strdup(ex->_rep_id());
#endif
}

CORBA::UnknownUserException *
corba_unknown_ex_narrow(CORBA::Exception * ex)
{
  return CORBA::UnknownUserException::_downcast(ex);
}

// Answer the repoid of the user exception contained in the exception
char *
corba_unknown_ex_get_repoid(CORBA::UnknownUserException * ex)
{
  // TODO: Fix memory leaks in this code
  return strdup(ex->exception().type()->id());
}

// Answer the user exception contained in the exception
CORBA::Any *
corba_unknown_ex_get_value(CORBA::UnknownUserException * ex,
						   CORBA::TypeCode_ptr tc)
{
  //
  // This is not documented in the CORBA specification.
  // The 'exception(tc)' call tries to unmarshall the Any
  // contained in Exception using the TypeCode provided by the user.
  // If it succeeds, it returns an Any. Otherwise, it asserts....
  //

#ifdef HASKELL_MICO_ORB
  // Does the TypeCode argument necessary ?
  return &(ex->exception(tc));
#else
  return &(ex->exception());
#endif
}

////////////////////////////////
//      SERVER REQUEST
////////////////////////////////

char *
corba_server_request_operation(CORBA::ServerRequest_ptr req)
{
  return strdup(req->operation());
}

void
corba_server_request_set_params(CORBA::ServerRequest_ptr req,
								CORBA::NVList_ptr args)
{
  req->arguments(args);
}

void
corba_server_request_set_result(CORBA::ServerRequest_ptr req,
								CORBA::Any * result)
{
  req->set_result(*result);
}

void
corba_server_request_set_exception(CORBA::ServerRequest_ptr req,
								   CORBA::Any * result)
{
  req->set_exception(*result);
}

CORBA::NVList_ptr
corba_server_request_get_args(CORBA::ServerRequest_ptr req)
{
#ifdef HASKELL_ORB_MICO
  // begin-mico-extension
  return req->arguments();
  // end-mico-extension
#else
  // TODO:
  assert(0);
#endif
}

////////////////////////////////
//       SERVER SIDE
////////////////////////////////

// The ObjectId type does not have a comparison operator, thus it can not
// be stored in STL maps.
bool
operator< (const PortableServer::ObjectId& o1, const PortableServer::ObjectId & o2);

// Mapping ObjectIds to repository IDs
static map<PortableServer::ObjectId, char *, less<PortableServer::ObjectId> > * repository_ids = NULL;

// The repository ID of the newly created CORBA Object
static char * the_repo_id;

// This haskell callback function will process the incoming invocations
static void (*corba_server_callback)(void * ps_ref, void * req_ptr, void * obj_ptr);

// A stable pointer to the PS structure passed to the above callback
static void * ps_ptr;

class MyServant : virtual public PortableServer::DynamicImplementation {
  
  virtual void invoke (CORBA::ServerRequest_ptr req) {
	// The _this() methods throws an exception under ORBacus :(
	corba_server_callback(ps_ptr, req, _this());
  }

  virtual CORBA::RepositoryId _primary_interface(const PortableServer::ObjectId& id, PortableServer::POA_ptr poa) {
	// According to the CORBA 2.2 spec, the ORB only calls this function
	// when serving a request. However, under MICO 2.3.0, the ORB calls
	// this function when the application calls 
	// PortableServer::activate_object()...

	// Since the primary_interface of each object is permanent, we
	// store it in an STL map
	if (repository_ids == NULL)
	  repository_ids = new map<PortableServer::ObjectId, char *, less<PortableServer::ObjectId> >;

	if (repository_ids->find(id) == repository_ids->end()) {
	  (*repository_ids)[id] = strdup(the_repo_id);
	  //	  delete[] the_repo_id;
	}

    return(CORBA::string_dup((*repository_ids)[id]));
  }
};

bool
operator< (const PortableServer::ObjectId& o1, const PortableServer::ObjectId & o2)
{
  if (o1.length() != o2.length()) {
    return (o1.length() < o2.length());
  }

  for (unsigned int i = 0; i < o1.length(); ++i)
	if (o1[i] < o2[i])
	  return(true);
  return(false);
}

CORBA::Object_ptr
corba_servant_create(char * repoid)
{
  // This is freed in MyServant::_primary_interface()
  the_repo_id = strdup(repoid);

  PortableServer::ObjectId * id = the_poa->activate_object(new MyServant);

  return the_poa->id_to_reference(*id);
}

void corba_server_run(void * callback_ptr, void * a_ps_ptr)
{
  (void*)corba_server_callback = callback_ptr;  
  ps_ptr = a_ps_ptr;

  PortableServer::POAManager_var mgr = the_poa->the_POAManager();
  mgr->activate();

  the_orb->run();
}

void corba_server_exit()
{
  the_orb->shutdown(false);
}

} // extern "C"  

  

// Local Variables:
// tab-width: 4
// End:



